#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "type.h"
#include "utils.h"
#include "random.h"

Coords *cities;

double TSPDist(ChromType c1, ChromType c2);
void RemoveCrossings(IPTR pj, int start, int end);
void Reverse(IPTR tmp, int lchrom, ChromType c1, ChromType c2);

double Eval(IPTR pj)
{
  int i;
  double sum;
  int start, end;
  /*
  start = Rnd(0, pj->chromLen - 1);
  end = Rnd(0, pj->chromLen - 1);
  if(start > end) SwapChromType(&start, &end);
  *******************/

  start = 0;
  end = pj->chromLen;

  if(Flip(0.05))
    RemoveCrossings(pj, start, end);
  //  printf("After Remove Crossings\n");

  sum = 0.0;
  for(i = 0; i < pj->chromLen - 1; i++){
    sum += TSPDist(pj->chrom[i], pj->chrom[i+1]);
  }
  sum += TSPDist(pj->chrom[i], pj->chrom[0]);
  return sum;
}


double TSPDist(ChromType c1, ChromType c2)
{
  double xd, yd;
  xd = (double) (cities[c1].x - cities[c2].x);
  yd = (double) (cities[c1].y - cities[c2].y);
  return rint(sqrt((xd * xd) + (yd * yd)));

}


void AppInitChrom(IPTR pj)
{
  Shuffle(pj->chrom, pj->chromLen);
  
  return;
}

void AppSkipline(FILE *fp)
{
  int ch; 
  while( (ch = fgetc(fp)) != '\n'){
    if(ch == EOF){
      fprintf(stderr, "AppSkipline: Premature end of file \n");
      exit(1);
    }
  }
}


void AppInit(char *appInfile, Population *p)
{
  int i, c, tmp;
  FILE *fp;
  cities = (Coords *) calloc ((size_t) p->chromLength, sizeof(Coords));
  if ((fp = fopen(appInfile, "r")) == NULL) {
    fprintf(stderr, "AppInit: Cannot open %s for reading\n", appInfile);
    exit(1);
  }
  while( isalpha(c = fgetc(fp))){ // skip header 
    AppSkipline(fp);
  }
  ungetc(c, fp);

  // read coords
  for(i = 0; i < p->chromLength; i++){
    fscanf(fp, "%d %f %f", &tmp, &(cities[i].x), &(cities[i].y));
  }
  //  for(i = 0; i < p->chromLength; i++){
  //    fprintf(stdout, "%d %f %f \n", i, (cities[i].x), (cities[i].y));
  //  }

}

void PhenoPrint(FILE *fp, IPTR pop, Population *p)
{
  int i;
  
  fprintf(fp, "%i ", (int) pop[p->maxi].objfunc); 
  for(i = 0; i < p->chromLength; i++){
    fprintf(fp, " %f %f", 
	    cities[pop[p->maxi].chrom[i]].x, 
	    cities[pop[p->maxi].chrom[i]].y); 

  } 
  fprintf(fp, " %f %f", 
	  cities[pop[p->maxi].chrom[0]].x, 
	  cities[pop[p->maxi].chrom[0]].y); 

  fprintf(fp, "\n"); 
}

void TourPrint(FILE *fp, IPTR pj, char *name)
{
  int i;
  fprintf(fp, "%s :", name); 
  for(i = 0; i < pj->chromLen; i++){
    fprintf(fp, "%i ", pj->chrom[i]);
  } 
  fprintf(fp, "\n"); 
}
  


void InitPhenoPrint(IPTR pj, char *fname, Population *p)
{
  FILE *fp;
  int i;

  if((fp = fopen(fname, "w")) == NULL){
    fprintf(stderr, "InitPhenoPrint: Cannot open %s for writing\n", fname);
    exit(1);
  }
  fprintf(fp, "%i \nTourLength ", (p->chromLength + 1) * 2 + 1);
  for(i = 0; i < p->chromLength; i++){
    fprintf(fp, "x y ");

  }
  fprintf(fp, "x y ");
  fprintf(fp, "\n");
  fclose(fp);
}

void RemoveCrossings(IPTR pj, int start, int end)
{
  int i,j;
  int lchrom;
  IPTR tmp;
  ChromType city1Next, city2Prev;
  
  lchrom = pj->chromLen;
  tmp = AllocateIndividuals(1, lchrom);

  IndividualCopy(pj, tmp);
  for(i = start; i < end; i++){
    for(j = (i+3)%lchrom; j != i; j = ((j+1)%lchrom)){
      city1Next = (i + 1) % lchrom;
      city2Prev = (j + lchrom - 1) % lchrom;

      if((TSPDist(tmp->chrom[i], tmp->chrom[city1Next])
	  + TSPDist(tmp->chrom[j], tmp->chrom[city2Prev])) > 
	 (TSPDist(tmp->chrom[i], tmp->chrom[city2Prev]) + 
	  TSPDist(tmp->chrom[j], tmp->chrom[city1Next]))) {

	Reverse(tmp, lchrom, city1Next, city2Prev);

	/*	if(city2Prev < i) {   
		  i = 0; 
		  break; 
		  }*/
	/*	j = (i + 2) % lchrom; */
      }
    }
  }
  IndividualCopy(tmp, pj);
  free(tmp->chrom);
  free(tmp->backup);
  free(tmp);
  return;
}

void Reverse(IPTR tmp, int lchrom, ChromType c1, ChromType c2)
{
  SwapChromType(&(tmp->chrom[c1]), &(tmp->chrom[c2]));
  /****
  do {
    SwapChromType(&(tmp->chrom[c1]), &(tmp->chrom[c2]));
    c2 = (c2 + lchrom - 1) % lchrom;
    c1 = (c1 + 1) % lchrom;
  } while (c2 != c1 && ((c1 + lchrom - 1 ) % lchrom) != c2);
  ****/
return;
}
  
  
  
